
/**
 *
 * @file riscos.h
 *
 * Part of the OpenJazz project
 *
 * @par Licence:
 * Copyright (c) 2015-2023 Carsten Teibes
 *
 * OpenJazz is distributed under the terms of
 * the GNU General Public License, version 2.0
 *
 */


#ifndef _RISCOS_H
#define _RISCOS_H

#ifdef __riscos__

void RISCOS_AddGamePaths();

#endif

#endif
