0 ICON "OpenJazz.ico"
1 VERSIONINFO
FILEVERSION 0,0,0,0
FILETYPE 0x00000001L {
  BLOCK "StringFileInfo" {
    BLOCK "040904b0" {
      VALUE "CompanyName", ""
      VALUE "ProductName", "OpenJazz"
      VALUE "ProductVersion", "2024.09.19.0"
      VALUE "FileDescription", "Jazz Jackrabbit™ engine reimplementation"
      VALUE "LegalCopyright", "Alister Thomson et al."
      VALUE "OriginalFilename", "OpenJazz.exe"
      VALUE "InternalName", "OpenJazz"
      VALUE "Comments", "http://www.alister.eu/jazz/oj/"
    }
  }
  BLOCK "VarFileInfo" {
    VALUE "Translation", 0x409, 0x4b0
  }
}
