
/**
 *
 * @file version.h
 *
 * Part of the OpenJazz project
 *
 * @par Licence:
 * Copyright (c) 2015-2023 Carsten Teibes
 *
 * OpenJazz is distributed under the terms of
 * the GNU General Public License, version 2.0
 *
 */


#ifndef OJ_VERSION_H
#define OJ_VERSION_H

#include "OpenJazz.h"

extern const char *oj_version;
extern const char *oj_date;

#endif
