
/**
 *
 * @file wii.h
 *
 * Part of the OpenJazz project
 *
 * @par Licence:
 * Copyright (c) 2015-2023 Carsten Teibes
 *
 * OpenJazz is distributed under the terms of
 * the GNU General Public License, version 2.0
 *
 */


#ifndef _WII_H
#define _WII_H

#ifdef __wii__

void WII_Init();

void WII_AddGamePaths();

#endif

#endif
